from __future__ import annotations

import numpy as np
import pytest

from tanabesugano.matrices import LigandFieldTheory


# Define a type alias for clarity
Float64Array = np.ndarray


@pytest.fixture
def ligand_field_theory():
    # Arrange
    # Create a fixture for the LigandFieldTheory instance
    return LigandFieldTheory(Dq=1000.0, B=800.0, C=4000.0)


@pytest.mark.parametrize(
    "matrix, expected_eigenvalues",
    [
        (
            np.array([[1, 0], [0, 1]], dtype=np.float64),
            np.array([1, 1], dtype=np.float64),
        ),
        (
            np.array([[2, 1], [1, 2]], dtype=np.float64),
            np.array([1, 3], dtype=np.float64),
        ),
        (
            np.array([[0, 0], [0, 0]], dtype=np.float64),
            np.array([0, 0], dtype=np.float64),
        ),
    ],
    ids=["identity_matrix", "symmetric_matrix", "zero_matrix"],
)
def test_eigensolver(ligand_field_theory, matrix, expected_eigenvalues):
    # Act
    eigenvalues = ligand_field_theory.eigensolver(matrix)

    # Assert
    np.testing.assert_array_almost_equal(eigenvalues, expected_eigenvalues)


@pytest.mark.parametrize(
    "diag_elements, off_diag_elements, expected_matrix",
    [
        ([1, 2], {(0, 1): 0.5}, np.array([[1, 0.5], [0.5, 2]], dtype=np.float64)),
        (
            [3, 4, 5],
            {(0, 2): 1.0, (1, 2): 0.5},
            np.array([[3, 0, 1], [0, 4, 0.5], [1, 0.5, 5]], dtype=np.float64),
        ),
        ([0, 0], {}, np.array([[0, 0], [0, 0]], dtype=np.float64)),
    ],
    ids=["simple_2x2", "complex_3x3", "zero_2x2"],
)
def test_construct_matrix(
    ligand_field_theory, diag_elements, off_diag_elements, expected_matrix,
):
    # Act
    matrix = ligand_field_theory.construct_matrix(diag_elements, off_diag_elements)

    # Assert
    np.testing.assert_array_almost_equal(matrix, expected_matrix)


def test_solver_not_implemented(ligand_field_theory):
    # Act & Assert
    with pytest.raises(
        NotImplementedError, match="Subclasses should implement this method.",
    ):
        ligand_field_theory.solver()


@pytest.mark.parametrize(
    "matrix",
    [
        np.array([[1, 2], [3, 4]], dtype=np.float64),  # Non-symmetric matrix
        np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64),  # Non-square matrix
    ],
    ids=["non_symmetric_matrix", "non_square_matrix"],
)
@pytest.mark.xfail(reason="Functionality not implemented yet")
def test_eigensolver_invalid_input(ligand_field_theory, matrix):
    # Act & Assert
    with pytest.raises(ValueError):
        ligand_field_theory.eigensolver(matrix)



from __future__ import annotations

from typing import List

import numpy as np

from tanabesugano import matrices
from tanabesugano import tools


class Batch:
    def __init__(
        self,
        Dq: List[float] = None,
        B: List[float] = None,
        C: List[float] = None,
        d_count: int = 5,
        slater: bool = False,
    ):
        if Dq is None:
            Dq = [4000.0, 4500.0, 10]
        if B is None:
            B = [400.0, 4500.0, 10]
        if C is None:
            C = [3600.0, 4000, 10]
        if len(Dq) != 3:
            raise KeyError(
                "The range of `Dq` is based on the three values: start, stop, steps!",
            )
        self.Dq = np.linspace(Dq[0], Dq[1], int(Dq[2]))  # Oh-crystalfield-splitting
        if len(B) != 3:
            raise KeyError(
                "The range of `B` is based on the three values: start, stop, steps!",
            )
        self.B = np.linspace(
            B[0],
            B[1],
            int(B[2]),
        )  # Racah-Parameter B in wavenumbers
        if len(C) != 3:
            raise KeyError(
                "The range of `C` is based on the three values: start, stop, steps!",
            )
        self.C = np.linspace(
            C[0],
            C[1],
            int(C[2]),
        )  # Racah-Parameter C in wavenumbers

        if slater:
            # Transformin Racah to Slater-Condon
            self.B, self.C = tools.racah(B, C)

        # self.delta_B = self.e_range / self.B

        self.d_count = d_count
        if self.d_count in {4, 5, 6}:
            self._size = 42
        if self.d_count in {3, 7}:
            self._size = 19
        if self.d_count in {2, 8}:
            self._size = 10
        self.result: List[dict] = []

    def calculation(self) -> None:
        """Is filling the self.result with the iTS states of over-iterated energy range."""
        for _Dq in self.Dq:
            for _B in self.B:
                for _C in self.C:
                    if self.d_count == 2:  # d2
                        states = matrices.d2(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )

                    elif self.d_count == 3:  # d3
                        states = matrices.d3(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    elif self.d_count == 4:  # d4
                        states = matrices.d4(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    elif self.d_count == 5:  # d5
                        states = matrices.d5(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    elif self.d_count == 6:  # d6
                        states = matrices.d6(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    elif self.d_count == 7:  # d7
                        states = matrices.d7(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    elif self.d_count == 8:  # d8
                        states = matrices.d8(Dq=_Dq, B=_B, C=_C).solver()
                        self.result.append(
                            {
                                "d_count": self.d_count,
                                "Dq": _Dq,
                                "B": _B,
                                "C": _C,
                                "states": states,
                            },
                        )
                    else:
                        raise ValueError("not a correct value!")

    @property
    def return_result(self) -> List[dict]:
        return self.result


if __name__ == "__main__":
    res = Batch()
    res.calculation()
    print(res.return_result)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Tuple
from typing import Union


if TYPE_CHECKING:
    import numpy as np


def racah(
    F2: Union[float, np.array],
    F4: Union[float, np.array],
) -> Union[Tuple[float, float], Tuple[np.array, np.array]]:
    """Transform the Slater-Condon-Parameter to Racah-Parameter.

    Re-calculating and normalization of the Slater-Condon-Parameter to Racah-Parameter
        eV will be converted to wavenumbers

    Args:
        F2 (Union[float, np.array]): Slater-Condon-Pramater `F2` as float-value or
            array.
        F4 (Union[float, np.array]): Slater-Condon-Pramater `F4` as float-value or
            array.

    Returns:
        Union[Tuple[float, float], Tuple[np.array, np.array]]: Return the racah
            parameters.

    """
    eVcm = 8065.54
    B = eVcm * (F2 / 49.0 - 5 / 441.0 * F4)
    C = eVcm * (35 / 441.0 * F4)
    return B, C



#!/usr/bin/env python

from __future__ import annotations

import argparse

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from prettytable import PrettyTable


try:
    import plotly.express as px
except ImportError:  # pragma: no cover
    px = None

from tanabesugano import __version__
from tanabesugano import matrices
from tanabesugano import tools


class CMDmain:
    def __init__(
        self,
        Dq: float = 4000.0,
        B: float = 400.0,
        C: float = 3600.0,
        nroots: int = 100,
        d_count: int = 5,
        slater: bool = False,
    ) -> None:
        """CMD Interface for Tanabe-Sugano-Diagram

        Parameters
        ----------
        Dq : float, optional
            Oh-crystalfield-splitting, by default 4000.0
        B : float, optional
            Racah-Parameter B in wavenumbers, by default 400.0
        C : float, optional
            Racah-Parameter C in wavenumbers, by default 3600.0
        nroots : int, optional
            Number of roots to calculate the TS-diagram , by default 100
        d_count : int, optional
            Electron count, by default 5
        slater : bool, optional
             Transforming from Racah to Slater-Condon, by default False

        """
        self.Dq = Dq
        self.B = B
        self.C = C

        if slater:
            self.B, self.C = tools.racah(B, C)
        self.nroot = nroots
        energy = np.linspace(0.0, self.Dq, nroots)

        self.d_count = d_count
        if self.d_count in {4, 5, 6}:
            self._size = 42
        if self.d_count in {3, 7}:
            self._size = 19
        if self.d_count in {2, 8}:
            self._size = 10
        self.result = np.zeros((self._size + 1, nroots))

        self.df = pd.DataFrame(
            {"Energy": energy, "delta_B": energy / self.B, "10Dq": energy * 10.0},
        )
        self.title_TS = (
            f"TS-diagram_d{self.d_count}_10Dq_{int(self.Dq * 10.0)}_"
            f"B_{int(self.B)}_C_{int(self.C)}"
        )
        self.title_DD = (
            f"DD-energies_d{self.d_count}_10Dq_{int(self.Dq * 10.0)}_"
            f"B_{int(self.B)}_C_{int(self.C)}"
        )

    def plot(self) -> None:
        # Figure one for classical Tanabe-Sugano-Diagram with B-dependency
        plt.figure(1)

        # Set Window Title

        plt.plot(
            self.df["delta_B"],
            # get the states from self.df
            self.df.drop(["Energy", "delta_B", "10Dq"], axis=1).to_numpy() / self.B,
            ls="--",
        )
        self.label_plot("Tanabe-Sugano-Diagram", "$E/B$", r"$\Delta/B$")
        # Figure one for Energy-Correlation-Diagram Dq-Energy versus State-Energy
        plt.figure(2)

        plt.plot(
            self.df["10Dq"],
            self.df.drop(["Energy", "delta_B", "10Dq"], axis=1).to_numpy(),
            ls="--",
        )
        self.label_plot(
            "DD excitations -Diagram",
            r"$dd-state-energy\,(1/cm)$",
            r"$10Dq\,(1/cm)$",
        )
        plt.show()

    def label_plot(self, arg0: str, arg1: str, arg2: str) -> None:
        """Labels the plot."""
        plt.title(arg0)
        plt.ylabel(arg1)
        plt.xlabel(arg2)

    def savetxt(self) -> None:
        pd.concat(
            [
                self.df["delta_B"],
                self.df.drop(["Energy", "delta_B", "10Dq"], axis=1) / self.B,
            ],
            axis=1,
        ).to_csv(Path(f"{self.title_TS}.csv"), index=False)

        pd.concat(
            [self.df["10Dq"], self.df.drop(["Energy", "delta_B", "10Dq"], axis=1)],
            axis=1,
        ).to_csv(Path(f"{self.title_DD}.csv"), index=False)

    def calculation(self) -> None:
        """Is filling the self.result with the iTS states of over-iterated energy range"""
        result = []
        for dq in self.df["Energy"]:
            if self.d_count == 2:  # d2
                result.append(
                    self.subsplit_states(
                        matrices.d2(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 3:  # d3
                result.append(
                    self.subsplit_states(
                        matrices.d3(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 4:  # d4
                result.append(
                    self.subsplit_states(
                        matrices.d4(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 5:  # d5
                result.append(
                    self.subsplit_states(
                        matrices.d5(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 6:  # d6
                result.append(
                    self.subsplit_states(
                        matrices.d6(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 7:  # d7
                result.append(
                    self.subsplit_states(
                        matrices.d7(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            elif self.d_count == 8:  # d8
                result.append(
                    self.subsplit_states(
                        matrices.d8(Dq=dq, B=self.B, C=self.C).solver(),
                    ),
                )
            else:
                msg = "The number of unpaired electrons should be between 2 and 8."
                raise ValueError(msg)

        # Transform list of dictionaries to dictionary of arrays
        result = {
            key: np.array([d[key] for d in result]).flatten() for key in result[0]
        }
        self.df = pd.concat([self.df, pd.DataFrame(result)], axis=1)

    @staticmethod
    def subsplit_states(states: dict) -> dict:
        """Subsplitting the states for a better overview."""
        rearranged_states = {}
        for key, value in states.items():
            if len(value) > 1:
                for i, _value in enumerate(value):
                    rearranged_states[f"{key}_{i}"] = np.array([_value])
            else:
                rearranged_states[key] = value
        return rearranged_states

    def ci_cut(self, dq_ci: float = None) -> None:
        """Extracting the atomic-termsymbols for a specific dq depending on the oxidation state"""
        if self.d_count == 2:  # d2
            states = matrices.d2(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 3:  # d3
            states = matrices.d3(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 4:  # d4
            states = matrices.d4(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 5:  # d5
            states = matrices.d5(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 6:  # d6
            states = matrices.d6(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 7:  # d7
            states = matrices.d7(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

        elif self.d_count == 8:  # d8
            states = matrices.d8(Dq=dq_ci / 10.0, B=self.B, C=self.C).solver()
            self.ts_print(states, dq_ci=dq_ci)

    def ts_print(self, states: dict, dq_ci: float = None) -> None:
        """Print the atomic-termsymbols

        Print the atomic-termsymbols for a specific dq depending on the oxidation state
        on the screen and save them as txt-file.

        Parameters
        ----------
        states : dict
            List of atomic-termsymbols for a specific oxidation state
        dq_ci : float, optional
            Specific crystalfield-splitting in Dq, by default None

        """
        count = 0
        cut = np.zeros(
            self._size + 1,
            dtype=[("state", np.str_, 7), ("cm", int), ("eV", float)],
        )
        for irreducible in states:
            for energy in states[irreducible]:
                cut["state"][count] = irreducible
                cut["cm"][count] = np.round(energy, 0).astype(int)
                cut["eV"][count] = np.round(energy * 0.00012, 4)

                count += 1

        results = np.sort(cut, order="eV")

        x = PrettyTable(results.dtype.names)
        for row in results:
            x.add_row(row)
        # Change some column alignments; default was 'c'
        x.field_names = ["State", "cm-", "eV"]
        x.align["state"] = "l"
        x.align["cm"] = "r"
        x.align["eV"] = "r"
        print(x)
        title = "TS_Cut_d%i_10Dq_%i_B_%i_C_%i.csv" % (
            self.d_count,
            dq_ci,
            self.B,
            self.C,
        )

        np.savetxt(
            title,
            results.T,
            delimiter=",",
            header="state,cm,eV",
            fmt=r"%s,%i,%.4f",
            # Remove # for comments
            comments="",
        )

    def interactive_plot(self) -> None:
        """Interactive plot for the tanabe-sugano-diagram."""
        if px is None:
            raise ImportError(
                "Plotly is not installed. Please install plotly "
                "with 'pip install tanabesugano[plotly]'!",
            )

        _col = self.df.drop(["Energy", "delta_B", "10Dq"], axis=1).columns
        _font = dict(family="Avant Garde, sans-serif", size=12, color="grey")
        _template = "plotly_white"
        _size = dict(
            autosize=False,
            width=800,
            height=800,
        )
        color_discrete_sequence = [
            px.colors.qualitative.Light24[int(i[0]) - 1] for i in _col
        ]

        _df = self.df.copy()

        fig_1 = px.line(
            _df,
            x="10Dq",
            y=_col,
            title="Energy-Correlation-Diagram",
            labels={
                "variable": "State",
                "value": "Energy (cm-1)",
                "10Dq": "10Dq (cm-1)",
            },
            color_discrete_sequence=color_discrete_sequence,
        )
        fig_1.update_layout(
            xaxis_title="10Dq (cm-1)",
            yaxis_title="dd-states (cm-1)",
            legend_title="State",
            template=_template,
            font=_font,
            **_size,
        )
        # Save as html-file
        fig_1.write_html(Path(f"{self.title_DD}.html"))

        # Apply / self.B to every column except for _col
        _df[_col] = _df[_col].div(self.B, axis=0)

        # Plot the tanabe-sugano-diagram
        fig_2 = px.line(
            _df,
            x="delta_B",
            y=_col,
            title="Tanabe-Sugano-Diagram",
            labels={
                "variable": "State",
                "value": " E / B",
                "delta_B": "Δ / B",
            },
            color_discrete_sequence=color_discrete_sequence,
        )
        fig_2.update_layout(
            xaxis_title="Δ / B",
            yaxis_title="E / B",
            legend_title="State",
            template=_template,
            font=_font,
            **_size,
        )
        # Save as html-file
        fig_2.write_html(Path(f"{self.title_TS}.html"))


def cmd_line() -> None:
    """Command line interface for tanabe-sugano."""
    description = (
        "A python-based Eigensolver for Tanabe-Sugano- & Energy-Correlation-Diagrams "
        "based on the original three proposed studies of *Yukito Tanabe and Satoru Sugano* "
        "for d3-d8 transition metal ions:\n"
        "For futher help, please use tanabe '--help'"
    )

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "-d",
        type=int,
        default=6,
        help="Number of unpaired electrons (default d5)",
    )
    parser.add_argument(
        "-Dq",
        type=float,
        default=25065.0,
        help="10Dq crystal field splitting (default 10Dq = 8065 cm-)",
    )
    parser.add_argument(
        "-cut",
        type=float,
        default=24000,
        help="10Dq crystal field splitting (default 10Dq = 8065 cm-)",
    )
    parser.add_argument(
        "-B",
        type=float,
        nargs=2,
        default=[1080.0, 1.0],
        help="Racah Parameter B and the corresponding "
        "reduction (default B = 860 cm- * 1.)",
    )
    parser.add_argument(
        "-C",
        type=float,
        nargs=2,
        default=[4773.0, 1.0],
        help="Racah Parameter C and the corresponding "
        "reduction (default C = 4.477*860 cm- * "
        "1.)",
    )
    parser.add_argument(
        "-n",
        type=int,
        default=500,
        help="Number of roots (default nroots = 500)",
    )
    parser.add_argument(
        "-ndisp",
        action="store_true",
        default=False,
        help="Plot TS-diagram (default = on)",
    )
    parser.add_argument(
        "-ntxt",
        action="store_true",
        default=False,
        help="Save TS-diagram and dd energies (default = on)",
    )
    parser.add_argument(
        "-slater",
        action="store_true",
        default=False,
        help="Using Slater-Condon F2,F4 parameter "
        "instead Racah-Parameter B,C (default = off)",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Print version number and exit",
    )
    parser.add_argument(
        "-html",
        action="store_true",
        default=False,
        help="Save TS-diagram and dd energies (default = off)",
    )

    args = parser.parse_args()

    tmm = CMDmain(
        Dq=args.Dq / 10.0,
        B=args.B[0] * args.B[1],
        C=args.C[0] * args.C[1],
        nroots=args.n,
        d_count=args.d,
        slater=args.slater,
    )
    tmm.calculation()

    if args.ndisp is not True:
        tmm.plot()
    if args.ntxt is not True:
        tmm.savetxt()
    if args.cut is not None:
        tmm.ci_cut(dq_ci=args.cut)
    if args.html:
        tmm.interactive_plot()



from __future__ import annotations

from typing import Any

from tanabesugano import cmd as frontapp


def test_frontapp():
    return frontapp.CMDmain(
        Dq=4000.0,
        B=400.0,
        C=3600.0,
        nroots=100,
        d_count=5,
    ).calculation()


def test_cmd(script_runner: Any) -> None:
    ret = script_runner.run("tanabesugano", "--help")
    assert ret.success



from __future__ import annotations
import stat

import numpy as np

from tanabesugano import matrices


def state_check(x) -> int:
    for i in np.linspace(0, 1500, 30):
        if x == 2:
            states = matrices.d2(Dq=i).solver()
            return len(states)
        if x == 3:
            states = matrices.d3(Dq=i).solver()
            return len(states)
        if x == 4:
            states = matrices.d4(Dq=i).solver()
            return len(states)
        if x == 5:
            states = matrices.d5(Dq=i).solver()
            return len(states)
        if x == 6:
            states = matrices.d6(Dq=i).solver()
            return len(states)
        if x == 7:
            states = matrices.d7(Dq=i).solver()
            return len(states)
        if x == 8:
            states = matrices.d8(Dq=i).solver()
            return len(states)

def test_answer_d2() -> None:
    assert state_check(2) == 7

def test_answer_d3() -> None:
    assert state_check(3) == 8


def test_answer_d4() -> None:
    assert state_check(4) == 12


def test_answer_d5() -> None:
    assert state_check(5) == 11


def test_answer_d6() -> None:
    assert state_check(6) == 12


def test_answer_d7() -> None:
    assert state_check(7) == 8



from __future__ import annotations

from tanabesugano.batch import Batch


def test_batch_small():
    res = Batch()
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d2():
    res = Batch(d_count=2)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d3():
    res = Batch(d_count=3)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d4():
    res = Batch(d_count=4)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d5():
    res = Batch(d_count=5)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d6():
    res = Batch(d_count=6)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d7():
    res = Batch(d_count=7)
    res.calculation()
    assert isinstance(res.result, list)


def test_batch_d8():
    res = Batch(d_count=8)
    res.calculation()
    assert isinstance(res.result, list)



from __future__ import annotations

from typing import Dict
from typing import List
from typing import Tuple


try:
    from typing import TypeAlias
except ImportError:
    from typing import Any as TypeAlias

import numpy as np

from numpy._typing._array_like import NDArray
from numpy.linalg import eigh


_sqrt2 = np.sqrt(2.0)
_sqrt3 = np.sqrt(3.0)
_sqrt6 = np.sqrt(6.0)

_2sqrt2 = _sqrt2 * 2.0
_2sqrt3 = _sqrt3 * 2.0
_3sqrt2 = _sqrt2 * 3.0
_3sqrt3 = _sqrt3 * 3.0
_3sqrt6 = _sqrt6 * 3.0

Float64Array: TypeAlias = NDArray[np.float64]


class LigandFieldTheory:
    """Parent class for ligand field theory configurations."""

    def __init__(self, Dq: float, B: float, C: float) -> None:
        """Initialize the configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        self.Dq = np.float64(Dq)
        self.B = np.float64(B)
        self.C = np.float64(C)

    def eigensolver(self, matrix: Float64Array) -> Float64Array:
        """Solve for the eigenvalues of the given matrix.

        Args:
            matrix (Float64Array): 2-dimensional square array representing the TS matrix
                of the ligand field Hamiltonian.

        Returns:
            Float64Array: 1-dimensional array of eigenvalues of the diagonalized ligand
                field Hamiltonian.

        """
        return eigh(matrix)[0]

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        msg = "Subclasses should implement this method."
        raise NotImplementedError(msg)

    def construct_matrix(
        self,
        diag_elements: List[float],
        off_diag_elements: Dict[Tuple[int, int], float],
    ) -> Float64Array:
        """Construct a symmetric matrix from diagonal and off-diagonal elements."""
        size = len(diag_elements)
        matrix = np.zeros((size, size))
        np.fill_diagonal(matrix, diag_elements)
        for (i, j), value in off_diag_elements.items():
            matrix[i, j] = value
            matrix[j, i] = value  # Assuming the matrix is symmetric
        return matrix


class d2(LigandFieldTheory):
    """Class representing the d2 configuration in ligand field theory."""

    def __init__(self, Dq: float = 0.0, B: float = 860.0, C: float = 3801.0) -> None:
        """Initialize the d2 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def A_1_1_states(self) -> Float64Array:
        """Calculate the A_1_1 states."""
        diag_elements = [
            -8 * self.Dq + 10 * self.B + 5 * self.C,
            +12 * self.Dq + 8 * self.B + 4 * self.C,
        ]
        off_diag_elements = {(0, 1): _sqrt6 * (2 * self.B + self.C)}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_1_states(self) -> Float64Array:
        """Calculate the E_1 states."""
        diag_elements = [-8 * self.Dq + self.B + 2 * self.C, +12 * self.Dq + 2 * self.C]
        off_diag_elements = {(0, 1): -_2sqrt3 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_2_states(self) -> Float64Array:
        """Calculate the T_1_2 states."""
        diag_elements = [-8 * self.Dq + self.B + 2 * self.C, +2 * self.Dq + 2 * self.C]
        off_diag_elements = {(0, 1): +_2sqrt3 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_3_1_states(self) -> Float64Array:
        """Calculate the T_3_1 states."""
        diag_elements = [-8 * self.Dq - 5 * self.B, +2 * self.Dq + 4 * self.B]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states
        GS = self.T_3_1_states()[0]

        T_1_1 = np.array([+2 * self.Dq + 4 * self.B + 2 * self.C]) - GS
        T_3_2 = np.array([+2 * self.Dq - 8 * self.B]) - GS
        A_3_2 = np.array([12 * self.Dq - 8 * self.B]) - GS

        # Ligand field dependent states
        A_1_1 = self.A_1_1_states() - GS
        E_1 = self.E_1_states() - GS
        T_1_2 = self.T_1_2_states() - GS
        T_3_1 = self.T_3_1_states() - GS

        return {
            "1_A_1": A_1_1,
            "1_E": E_1,
            "1_T_3": T_1_2,
            "3_T_1": T_3_1,
            "1_T_1": T_1_1,
            "3_T_2": T_3_2,
            "3_A_2": A_3_2,
        }


class d3(LigandFieldTheory):
    """Class representing the d3 configuration in ligand field theory."""

    def __init__(self, Dq: float = 0.0, B: float = 918.0, C: float = 4133.0) -> None:
        """Initialize the d3 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def T_2_2_states(self) -> Float64Array:
        """Calculate the T_2_2 states."""
        diag_elements = [
            -12 * self.Dq + 5 * self.C,
            -2 * self.Dq - 6 * self.B + 3 * self.C,
            -2 * self.Dq + 4 * self.B + 3 * self.C,
            +8 * self.Dq + 6 * self.B + 5 * self.C,
            +8 * self.Dq - 2 * self.B + 3 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_3sqrt3 * self.B,
            (0, 2): -5 * _sqrt3 * self.B,
            (0, 3): 4 * self.B + 2 * self.C,
            (0, 4): 2 * self.B,
            (1, 2): 3 * self.B,
            (1, 3): -_3sqrt3 * self.B,
            (1, 4): -_3sqrt3 * self.B,
            (2, 3): -_sqrt3 * self.B,
            (2, 4): +_sqrt3 * self.B,
            (3, 4): 10 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_2_1_states(self) -> Float64Array:
        """Calculate the T_2_1 states."""
        diag_elements = [
            -12 * self.Dq - 6 * self.B + 3 * self.C,
            -2 * self.Dq + 3 * self.C,
            -2 * self.Dq - 6 * self.B + 3 * self.C,
            +8 * self.Dq - 6 * self.B + 3 * self.C,
            +8 * self.Dq - 2 * self.B + 3 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -3 * self.B,
            (0, 2): +3 * self.B,
            (0, 3): 0.0,
            (0, 4): -_2sqrt3 * self.B,
            (1, 2): -3 * self.B,
            (1, 3): +3 * self.B,
            (1, 4): _3sqrt3 * self.B,
            (2, 3): -3 * self.B,
            (2, 4): -_sqrt3 * self.B,
            (3, 4): _2sqrt3 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_2_states(self) -> Float64Array:
        """Calculate the E_2 states."""
        diag_elements = [
            -12 * self.Dq - 6 * self.B + 3 * self.C,
            -2 * self.Dq + 8 * self.B + 6 * self.C,
            -2 * self.Dq - 1 * self.B + 3 * self.C,
            +18 * self.Dq - 8 * self.B + 4 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -6 * _sqrt2 * self.B,
            (0, 2): -_3sqrt2 * self.B,
            (0, 3): 0.0,
            (1, 2): 10 * self.B,
            (1, 3): +_sqrt3 * (2 * self.B + self.C),
            (2, 3): _2sqrt3 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_4_1_states(self) -> Float64Array:
        """Calculate the T_4_1 states."""
        diag_elements = [-2 * self.Dq - 3 * self.B, +8 * self.Dq - 12 * self.B]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states
        GS = np.array([-12 * self.Dq - 15 * self.B])

        A_4_2 = np.array([0], dtype=np.float64)
        T_4_2 = np.array([-2 * self.Dq - 15 * self.B]) - GS

        A_2_1 = np.array([-2 * self.Dq - 11 * self.B + 3 * self.C]) - GS
        A_2_2 = np.array([-2 * self.Dq + 9 * self.B + 3 * self.C]) - GS

        # Ligand field dependent states
        T_2_2 = self.T_2_2_states() - GS
        T_2_1 = self.T_2_1_states() - GS
        E_2 = self.E_2_states() - GS
        T_4_1 = self.T_4_1_states() - GS

        return {
            "2_T_2": T_2_2,
            "2_T_1": T_2_1,
            "2_E": E_2,
            "4_T_1": T_4_1,
            "4_A_2": A_4_2,
            "4_T_2": T_4_2,
            "2_A_1": A_2_1,
            "2_A_2": A_2_2,
        }


class d4(LigandFieldTheory):
    """Class representing the d4 configuration in ligand field theory."""

    def __init__(self, Dq: float = 0.0, B: float = 965.0, C: float = 4449.0) -> None:
        """Initialize the d4 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def T_3_1_states(self) -> Float64Array:
        """Calculate the T_3_1 states."""
        diag_elements = [
            -16 * self.Dq - 15 * self.B + 5 * self.C,
            -6 * self.Dq - 11 * self.B + 4 * self.C,
            -6 * self.Dq - 3 * self.B + 6 * self.C,
            +4 * self.Dq - self.B + 6 * self.C,
            +4 * self.Dq - 9 * self.B + 4 * self.C,
            +4 * self.Dq - 11 * self.B + 4 * self.C,
            +14 * self.Dq - 16 * self.B + 5 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_sqrt6 * self.B,
            (0, 2): -_3sqrt2 * self.B,
            (0, 3): _sqrt2 * (2 * self.B + self.C),
            (0, 4): -_2sqrt2 * self.B,
            (0, 5): 0.0,
            (0, 6): 0.0,
            (1, 2): 5 * _sqrt3 * self.B,
            (1, 3): _sqrt3 * self.B,
            (1, 4): -_sqrt3 * self.B,
            (1, 5): 3 * self.B,
            (1, 6): _sqrt6 * self.B,
            (2, 3): -3 * self.B,
            (2, 4): -3 * self.B,
            (2, 5): 5 * _sqrt3 * self.B,
            (2, 6): _sqrt2 * (self.B + self.C),
            (3, 4): -10 * self.B,
            (3, 5): 0.0,
            (3, 6): _3sqrt2 * self.B,
            (4, 5): -2 * _sqrt3 * self.B,
            (4, 6): -_3sqrt2 * self.B,
            (5, 6): _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_2_states(self) -> Float64Array:
        """Calculate the T_1_2 states."""
        diag_elements = [
            -16 * self.Dq - 9 * self.B + 7 * self.C,
            -6 * self.Dq - 9 * self.B + 6 * self.C,
            -6 * self.Dq + 3 * self.B + 8 * self.C,
            +4 * self.Dq - 9 * self.B + 6 * self.C,
            +4 * self.Dq - 3 * self.B + 6 * self.C,
            +4 * self.Dq + 5 * self.B + 8 * self.C,
            +14 * self.Dq + 7 * self.C,
        ]
        off_diag_elements = {
            (0, 1): _3sqrt2 * self.B,
            (0, 2): -5 * _sqrt6 * self.B,
            (0, 3): 0.0,
            (0, 4): -_2sqrt2 * self.B,
            (0, 5): _sqrt2 * (2 * self.B + self.C),
            (0, 6): 0.0,
            (1, 2): -5 * _sqrt3 * self.B,
            (1, 3): 3 * self.B,
            (1, 4): -3 * self.B,
            (1, 5): -3 * self.B,
            (1, 6): -_sqrt6 * self.B,
            (2, 3): -3 * _sqrt3 * self.B,
            (2, 4): 5 * _sqrt3 * self.B,
            (2, 5): -5 * _sqrt3 * self.B,
            (2, 6): _sqrt2 * (3 * self.B + self.C),
            (3, 4): -6 * self.B,
            (3, 5): 0.0,
            (3, 6): -_3sqrt6 * self.B,
            (4, 5): -10 * self.B,
            (4, 6): _sqrt6 * self.B,
            (5, 6): _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_1_1_states(self) -> Float64Array:
        """Calculate the A_1_1 states."""
        diag_elements = [
            -16 * self.Dq + 10 * self.C,
            -6 * self.Dq + 6 * self.C,
            +4 * self.Dq + 14 * self.B + 11 * self.C,
            +4 * self.Dq - 3 * self.B + 6 * self.C,
            +24 * self.Dq - 16 * self.B + 8 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -12 * _sqrt2 * self.B,
            (0, 2): _sqrt2 * (4 * self.B + 2 * self.C),
            (0, 3): _2sqrt2 * self.B,
            (0, 4): 0.0,
            (1, 2): -12 * self.B,
            (1, 3): -6 * self.B,
            (1, 4): 0.0,
            (2, 3): 20 * self.B,
            (2, 4): _sqrt6 * (2 * self.B + self.C),
            (3, 4): 2 * _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_1_1_states(self) -> Float64Array:
        """Calculate the E_1_1 states."""
        diag_elements = [
            -16 * self.Dq - 9 * self.B + 7 * self.C,
            -6 * self.Dq - 6 * self.B + 6 * self.C,
            +4 * self.Dq + 5 * self.B + 8 * self.C,
            +4 * self.Dq + 6 * self.B + 9 * self.C,
            +4 * self.Dq - 3 * self.B + 6 * self.C,
        ]
        off_diag_elements = {
            (0, 1): 6 * self.B,
            (0, 2): _sqrt2 * (2 * self.B + self.C),
            (0, 3): -2 * self.B,
            (0, 4): -4 * self.B,
            (1, 2): -_3sqrt2 * self.B,
            (1, 3): -12 * self.B,
            (1, 4): 0.0,
            (2, 3): 10 * _sqrt2 * self.B,
            (2, 4): -10 * _sqrt2 * self.B,
            (3, 4): 0.0,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_3_2_states(self) -> Float64Array:
        """Calculate the T_3_2 states."""
        diag_elements = [
            -6 * self.Dq - 9 * self.B + 4 * self.C,
            -6 * self.Dq - 5 * self.B + 6 * self.C,
            +4 * self.Dq - 13 * self.B + 4 * self.C,
            +4 * self.Dq - 9 * self.B + 4 * self.C,
            +14 * self.Dq - 8 * self.B + 5 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -5 * _sqrt3 * self.B,
            (0, 2): _sqrt6 * self.B,
            (0, 3): _sqrt3 * self.B,
            (0, 4): -_sqrt6 * self.B,
            (1, 2): -_3sqrt2 * self.B,
            (1, 3): 3 * self.B,
            (1, 4): _sqrt2 * (3 * self.B + self.C),
            (2, 3): -2 * _sqrt2 * self.B,
            (2, 4): -6 * self.B,
            (3, 4): _3sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_1_states(self) -> Float64Array:
        """Calculate the T_1_1 states."""
        diag_elements = [
            -6 * self.Dq - 3 * self.B + 6 * self.C,
            -6 * self.Dq - 3 * self.B + 8 * self.C,
            +4 * self.Dq - 3 * self.B + 6 * self.C,
            +14 * self.Dq - 16 * self.B + 7 * self.C,
        ]
        off_diag_elements = {
            (0, 1): 5 * _sqrt3 * self.B,
            (0, 2): 3 * self.B,
            (0, 3): _sqrt6 * self.B,
            (1, 2): -5 * _sqrt3 * self.B,
            (1, 3): _sqrt2 * (self.B + self.C),
            (2, 3): -_sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_3_1_states(self) -> Float64Array:
        """Calculate the E_3_1 states."""
        diag_elements = [
            -6 * self.Dq - 13 * self.B + 4 * self.C,
            -6 * self.Dq - 10 * self.B + 4 * self.C,
            +4 * self.Dq - 11 * self.B + 4 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -4 * self.B,
            (0, 2): 0.0,
            (1, 2): -_3sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_3_2_states(self) -> Float64Array:
        """Calculate the A_3_2 states."""
        diag_elements = [
            -6 * self.Dq - 8 * self.B + 4 * self.C,
            +4 * self.Dq - 2 * self.B + 7 * self.C,
        ]
        off_diag_elements = {(0, 1): -12 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_1_2_states(self) -> Float64Array:
        """Calculate the A_1_2 states."""
        diag_elements = [
            -6 * self.Dq - 12 * self.B + 6 * self.C,
            +4 * self.Dq - 3 * self.B + 6 * self.C,
        ]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states
        GS = np.array([-6 * self.Dq - 21 * self.B])

        E_5_1 = np.array([0], dtype=np.float64)
        T_5_2 = np.array([4 * self.Dq - 21 * self.B]) - GS

        A_3_1 = np.array([-6 * self.Dq - 12 * self.B + 4 * self.C]) - GS

        # Ligand field dependent states
        T_1_2 = self.T_1_2_states() - GS
        T_3_1 = self.T_3_1_states() - GS
        A_1_1 = self.A_1_1_states() - GS
        E_1_1 = self.E_1_1_states() - GS
        T_3_2 = self.T_3_2_states() - GS
        T_1_1 = self.T_1_1_states() - GS
        E_3_1 = self.E_3_1_states() - GS
        A_3_2 = self.A_3_2_states() - GS
        A_1_2 = self.A_1_2_states() - GS

        if T_3_1[0] <= 0:
            T_1_2 -= T_3_1[0]
            A_1_1 -= T_3_1[0]
            E_1_1 -= T_3_1[0]
            T_3_2 -= T_3_1[0]
            T_1_1 -= T_3_1[0]
            E_3_1 -= T_3_1[0]
            A_3_2 -= T_3_1[0]
            A_1_2 -= T_3_1[0]
            E_5_1 -= T_3_1[0]
            T_5_2 -= T_3_1[0]
            A_3_1 -= T_3_1[0]
            T_3_1 -= T_3_1[0]

        return {
            "3_T_1": T_3_1,
            "1_T_2": T_1_2,
            "1_A_1": A_1_1,
            "1_E_1": E_1_1,
            "3_T_2": T_3_2,
            "1_T_1": T_1_1,
            "3_E_1": E_3_1,
            "3_A_2": A_3_2,
            "1_A_2": A_1_2,
            "5_E_1": E_5_1,
            "5_T_2": T_5_2,
            "3_A_1": A_3_1,
        }


class d5(LigandFieldTheory):
    """Class representing the d5 configuration in ligand field theory."""

    def __init__(self, Dq: float = 0.0, B: float = 860.0, C: float = 3850.0) -> None:
        """Initialize the d5 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def T_2_2_states(self) -> Float64Array:
        """Calculate the T_2_2 states."""
        diag_elements = [
            -20 * self.Dq - 20 * self.B + 10 * self.C,
            -10 * self.Dq - 8 * self.B + 9 * self.C,
            -10 * self.Dq - 18 * self.B + 9 * self.C,
            -16 * self.B + 8 * self.C,
            -12 * self.B + 8 * self.C,
            +2 * self.B + 12 * self.C,
            -6 * self.B + 10 * self.C,
            +10 * self.Dq - 18 * self.B + 9 * self.C,
            +10 * self.Dq - 8 * self.B + 9 * self.C,
            +20 * self.Dq - 20 * self.B + 10 * self.C,
        ]
        off_diag_elements = {
            (0, 1): _3sqrt6 * self.B,
            (0, 2): _sqrt6 * self.B,
            (0, 3): 0.0,
            (0, 4): -2 * _sqrt3 * self.B,
            (0, 5): 4 * self.B + 2 * self.C,
            (0, 6): 2 * self.B,
            (0, 7): 0.0,
            (0, 8): 0.0,
            (0, 9): 0.0,
            (1, 2): 3 * self.B,
            (1, 3): _sqrt6 / 2.0 * self.B,
            (1, 4): -_3sqrt2 / 2.0 * self.B,
            (1, 5): _3sqrt6 / 2.0 * self.B,
            (1, 6): _3sqrt6 / 2.0 * self.B,
            (1, 7): 0.0,
            (1, 8): 4 * self.B + self.C,
            (1, 9): 0.0,
            (2, 3): _3sqrt6 / 2.0 * self.B,
            (2, 4): -_3sqrt2 / 2.0 * self.B,
            (2, 5): +5 * _sqrt6 / 2.0 * self.B,
            (2, 6): -5 * _sqrt6 / 2.0 * self.B,
            (2, 7): self.C,
            (2, 8): 0.0,
            (2, 9): 0.0,
            (3, 4): 2 * _sqrt3 * self.B,
            (3, 5): 0.0,
            (3, 6): 0.0,
            (3, 7): -_3sqrt6 / 2.0 * self.B,
            (3, 8): -_sqrt6 / 2.0 * self.B,
            (3, 9): 0.0,
            (4, 5): -10 * _sqrt3 * self.B,
            (4, 6): 0.0,
            (4, 7): _3sqrt2 / 2.0 * self.B,
            (4, 8): _3sqrt2 / 2.0 * self.B,
            (4, 9): -2 * _sqrt3 * self.B,
            (5, 6): 0.0,
            (5, 7): -5 * _sqrt6 / 2.0 * self.B,
            (5, 8): -_3sqrt6 / 2.0 * self.B,
            (5, 9): 4 * self.B + 2 * self.C,
            (6, 7): -5 * _sqrt6 / 2.0 * self.B,
            (6, 8): _3sqrt6 / 2.0 * self.B,
            (6, 9): -2.0 * self.B,
            (7, 8): 3 * self.B,
            (7, 9): -_sqrt6 * self.B,
            (8, 9): -_3sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_2_1_states(self) -> Float64Array:
        """Calculate the T_2_1 states."""
        diag_elements = [
            -10 * self.Dq - 22 * self.B + 9 * self.C,
            -10 * self.Dq - 8 * self.B + 9 * self.C,
            -4 * self.B + 10 * self.C,
            -12 * self.B + 8 * self.C,
            -10 * self.B + 10 * self.C,
            -6 * self.B + 10 * self.C,
            +10 * self.Dq - 8 * self.B + 9 * self.C,
            +10 * self.Dq - 22 * self.B + 9 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -3 * self.B,
            (0, 2): -_3sqrt2 / 2.0 * self.B,
            (0, 3): _3sqrt2 / 2.0 * self.B,
            (0, 4): -_3sqrt2 / 2.0 * self.B,
            (0, 5): -_3sqrt6 / 2.0 * self.B,
            (0, 6): 0.0,
            (0, 7): self.C,
            (1, 2): _3sqrt2 / 2.0 * self.B,
            (1, 3): _3sqrt2 / 2.0 * self.B,
            (1, 4): 15 * _sqrt2 / 2.0 * self.B,
            (1, 5): 5 * _sqrt6 / 2.0 * self.B,
            (1, 6): 4 * self.B + self.C,
            (1, 7): 0.0,
            (2, 3): 0.0,
            (2, 4): 0.0,
            (2, 5): 10 * _sqrt3 * self.B,
            (2, 6): _3sqrt2 / 2.0 * self.B,
            (2, 7): -_3sqrt2 / 2.0 * self.B,
            (3, 4): 0.0,
            (3, 5): 0.0,
            (3, 6): -_3sqrt2 / 2.0 * self.B,
            (3, 7): -_3sqrt2 / 2.0 * self.B,
            (4, 5): 2 * _sqrt3 * self.B,
            (4, 6): 15 * _sqrt2 / 2.0 * self.B,
            (4, 7): -_3sqrt2 / 2.0 * self.B,
            (5, 6): 5 * _sqrt6 / 2.0 * self.B,
            (5, 7): -_3sqrt6 / 2.0 * self.B,
            (6, 7): -3 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_2_states(self) -> Float64Array:
        """Calculate the E_2 states."""
        diag_elements = [
            -10 * self.Dq - 4 * self.B + 12 * self.C,
            -10 * self.Dq - 13 * self.B + 9 * self.C,
            -4 * self.B + 10 * self.C,
            -16 * self.B + 8 * self.C,
            -12 * self.B + 8 * self.C,
            +10 * self.Dq - 13 * self.B + 9 * self.C,
            +10 * self.Dq - 4 * self.B + 12 * self.C,
        ]
        off_diag_elements = {
            (0, 1): 10 * self.B,
            (0, 2): 6 * self.B,
            (0, 3): 6 * _sqrt3 * self.B,
            (0, 4): 6 * _sqrt2 * self.B,
            (0, 5): -2 * self.B,
            (0, 6): 4 * self.B + 2 * self.C,
            (1, 2): -3 * self.B,
            (1, 3): 3 * _sqrt3 * self.B,
            (1, 4): 0.0,
            (1, 5): 2 * self.B + self.C,
            (1, 6): 2 * self.B,
            (2, 3): 0.0,
            (2, 4): 0.0,
            (2, 5): -3 * self.B,
            (2, 6): -6 * self.B,
            (3, 4): 2 * _sqrt6 * self.B,
            (3, 5): -3 * _sqrt3 * self.B,
            (3, 6): 6 * _sqrt3 * self.B,
            (4, 5): 0.0,
            (4, 6): 6 * _sqrt2 * self.B,
            (5, 6): -10 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_2_1_states(self) -> Float64Array:
        """Calculate the A_2_1 states."""
        diag_elements = [
            -10 * self.Dq - 3 * self.B + 9 * self.C,
            -12 * self.B + 8 * self.C,
            -19 * self.B + 8 * self.C,
            +10 * self.Dq - 3 * self.B + 9 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_3sqrt2 * self.B,
            (0, 2): 0.0,
            (0, 3): 6 * self.B + self.C,
            (1, 2): -4 * _sqrt3 * self.B,
            (1, 3): _3sqrt2 * self.B,
            (2, 3): 0.0,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_2_2_states(self) -> Float64Array:
        """Calculate the A_2_2 states."""
        diag_elements = [
            -10 * self.Dq - 23 * self.B + 9 * self.C,
            -12 * self.B + 8 * self.C,
            +10 * self.Dq - 23 * self.B + 9 * self.C,
        ]
        off_diag_elements = {
            (0, 1): _3sqrt2 * self.B,
            (0, 2): -2 * self.B + self.C,
            (1, 2): -_3sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_4_1_states(self) -> Float64Array:
        """Calculate the T_4_1 states."""
        diag_elements = [
            -10 * self.Dq - 25 * self.B + 6 * self.C,
            -16 * self.B + 7 * self.C,
            10 * self.Dq - 25 * self.B + 6 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_3sqrt2 * self.B,
            (0, 2): self.C,
            (1, 2): -_3sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_4_2_states(self) -> Float64Array:
        """Calculate the T_4_2 states."""
        diag_elements = [
            -10 * self.Dq - 17 * self.B + 6 * self.C,
            -22 * self.B + 5 * self.C,
            +10 * self.Dq - 17 * self.B + 6 * self.C,
        ]
        off_diag_elements = {
            (0, 1): _sqrt6 * self.B,
            (0, 2): +4 * self.B + self.C,
            (1, 2): -_sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_4_states(self) -> Float64Array:
        """Calculate the E_4 states."""
        diag_elements = [-22 * self.B + 5 * self.C, -21 * self.B + 5 * self.C]
        off_diag_elements = {(0, 1): -2 * _sqrt3 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states
        GS = -35 * self.B

        A_6_1 = np.array(
            [0.0],
            dtype=float,
        )  # Starting value is -35. * B, but has to set to zero per definition
        E_4 = self.E_4_states() - GS
        A_4_1 = np.array([-25 * self.B + 5 * self.C]) - GS
        A_4_2 = np.array([-13 * self.B + 7 * self.C]) - GS

        # Ligandfield dependent
        T_2_2 = self.T_2_2_states() - GS
        T_2_1 = self.T_2_1_states() - GS
        E_2 = self.E_2_states() - GS
        A_2_1 = self.A_2_1_states() - GS
        A_2_2 = self.A_2_2_states() - GS
        T_4_1 = self.T_4_1_states() - GS
        T_4_2 = self.T_4_2_states() - GS

        if T_2_2[0] <= 0:
            A_6_1 -= T_2_2[0]
            E_4 -= T_2_2[0]
            A_4_1 -= T_2_2[0]
            A_4_2 -= T_2_2[0]
            T_2_1 -= T_2_2[0]
            E_2 -= T_2_2[0]
            A_2_1 -= T_2_2[0]
            A_2_2 -= T_2_2[0]
            T_4_1 -= T_2_2[0]
            T_4_2 -= T_2_2[0]
            # Finally create new ligand field independent state
            T_2_2 -= T_2_2[0]

        return {
            "2_T_2": T_2_2,
            "2_T_1": T_2_1,
            "2_E": E_2,
            "2_A_1": A_2_1,
            "2_A_2": A_2_2,
            "4_T_1": T_4_1,
            "4_T_2": T_4_2,
            "4_E": E_4,
            "6_A_1": A_6_1,
            "4_A_1": A_4_1,
            "4_A_2": A_4_2,
        }


class d6(LigandFieldTheory):
    """Class representing the d6 configuration in ligand field theory."""

    def __init__(self, Dq: float = 0.0, B: float = 1065.0, C: float = 5120.0) -> None:
        """Initialize the d6 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def T_3_1_states(self) -> Float64Array:
        """Calculate the T_3_1 states."""
        diag_elements = [
            +16 * self.Dq - 15 * self.B + 5 * self.C,
            +6 * self.Dq - 11 * self.B + 4 * self.C,
            +6 * self.Dq - 3 * self.B + 6 * self.C,
            -4 * self.Dq - self.B + 6 * self.C,
            -4 * self.Dq - 9 * self.B + 4 * self.C,
            -4 * self.Dq - 11 * self.B + 4 * self.C,
            -14 * self.Dq - 16 * self.B + 5 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_sqrt6 * self.B,
            (0, 2): -_3sqrt2 * self.B,
            (0, 3): _sqrt2 * (2 * self.B + self.C),
            (0, 4): -_2sqrt2 * self.B,
            (0, 5): 0.0,
            (0, 6): 0.0,
            (1, 2): 5 * _sqrt3 * self.B,
            (1, 3): _sqrt3 * self.B,
            (1, 4): -_sqrt3 * self.B,
            (1, 5): 3 * self.B,
            (1, 6): _sqrt6 * self.B,
            (2, 3): -3 * self.B,
            (2, 4): -3 * self.B,
            (2, 5): 5 * _sqrt3 * self.B,
            (2, 6): _sqrt2 * (self.B + self.C),
            (3, 4): -10 * self.B,
            (3, 5): 0.0,
            (3, 6): _3sqrt2 * self.B,
            (4, 5): -2 * _sqrt3 * self.B,
            (4, 6): -_3sqrt2 * self.B,
            (5, 6): _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_2_states(self) -> Float64Array:
        """Calculate the T_1_2 states."""
        diag_elements = [
            +16 * self.Dq - 9 * self.B + 7 * self.C,
            +6 * self.Dq - 9 * self.B + 6 * self.C,
            +6 * self.Dq + 3 * self.B + 8 * self.C,
            -4 * self.Dq - 9 * self.B + 6 * self.C,
            -4 * self.Dq - 3 * self.B + 6 * self.C,
            -4 * self.Dq + 5 * self.B + 8 * self.C,
            -14 * self.Dq + 7 * self.C,
        ]
        off_diag_elements = {
            (0, 1): _3sqrt2 * self.B,
            (0, 2): -5 * _sqrt6 * self.B,
            (0, 3): 0.0,
            (0, 4): -_2sqrt2 * self.B,
            (0, 5): _sqrt2 * (2 * self.B + self.C),
            (0, 6): 0.0,
            (1, 2): -5 * _sqrt3 * self.B,
            (1, 3): 3 * self.B,
            (1, 4): -3 * self.B,
            (1, 5): -3 * self.B,
            (1, 6): -_sqrt6 * self.B,
            (2, 3): -3 * _sqrt3 * self.B,
            (2, 4): 5 * _sqrt3 * self.B,
            (2, 5): -5 * _sqrt3 * self.B,
            (2, 6): _sqrt2 * (3 * self.B + self.C),
            (3, 4): -6 * self.B,
            (3, 5): 0.0,
            (3, 6): -_3sqrt6 * self.B,
            (4, 5): -10 * self.B,
            (4, 6): _sqrt6 * self.B,
            (5, 6): _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_1_1_states(self) -> Float64Array:
        """Calculate the A_1_1 states."""
        diag_elements = [
            +16 * self.Dq + 10 * self.C,
            +6 * self.Dq + 6 * self.C,
            -4 * self.Dq + 14 * self.B + 11 * self.C,
            -4 * self.Dq - 3 * self.B + 6 * self.C,
            -24 * self.Dq - 16 * self.B + 8 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -12 * _sqrt2 * self.B,
            (0, 2): _sqrt2 * (4 * self.B + 2 * self.C),
            (0, 3): _2sqrt2 * self.B,
            (0, 4): 0.0,
            (1, 2): -12 * self.B,
            (1, 3): -6 * self.B,
            (1, 4): 0.0,
            (2, 3): 20 * self.B,
            (2, 4): _sqrt6 * (2 * self.B + self.C),
            (3, 4): 2 * _sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_1_1_states(self) -> Float64Array:
        """Calculate the E_1_1 states."""
        diag_elements = [
            +16 * self.Dq - 9 * self.B + 7 * self.C,
            +6 * self.Dq - 6 * self.B + 6 * self.C,
            -4 * self.Dq + 5 * self.B + 8 * self.C,
            -4 * self.Dq + 6 * self.B + 9 * self.C,
            -4 * self.Dq - 3 * self.B + 6 * self.C,
        ]
        off_diag_elements = {
            (0, 1): 6 * self.B,
            (0, 2): _sqrt2 * (2 * self.B + self.C),
            (0, 3): -2 * self.B,
            (0, 4): -4 * self.B,
            (1, 2): -_3sqrt2 * self.B,
            (1, 3): -12 * self.B,
            (1, 4): 0.0,
            (2, 3): 10 * _sqrt2 * self.B,
            (2, 4): -10 * _sqrt2 * self.B,
            (3, 4): 0.0,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_3_2_states(self) -> Float64Array:
        """Calculate the T_3_2 states."""
        diag_elements = [
            +6 * self.Dq - 9 * self.B + 4 * self.C,
            +6 * self.Dq - 5 * self.B + 6 * self.C,
            -4 * self.Dq - 13 * self.B + 4 * self.C,
            -4 * self.Dq - 9 * self.B + 4 * self.C,
            -14 * self.Dq - 8 * self.B + 5 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -5 * _sqrt3 * self.B,
            (0, 2): _sqrt6 * self.B,
            (0, 3): _sqrt3 * self.B,
            (0, 4): -_sqrt6 * self.B,
            (1, 2): -_3sqrt2 * self.B,
            (1, 3): 3 * self.B,
            (1, 4): _sqrt2 * (3 * self.B + self.C),
            (2, 3): -2 * _sqrt2 * self.B,
            (2, 4): -6 * self.B,
            (3, 4): 3 * _sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_1_states(self) -> Float64Array:
        """Calculate the T_1_1 states."""
        diag_elements = [
            +6 * self.Dq - 3 * self.B + 6 * self.C,
            +6 * self.Dq - 3 * self.B + 8 * self.C,
            -4 * self.Dq - 3 * self.B + 6 * self.C,
            -14 * self.Dq - 16 * self.B + 7 * self.C,
        ]
        off_diag_elements = {
            (0, 1): 5 * _sqrt3 * self.B,
            (0, 2): 3 * self.B,
            (0, 3): _sqrt6 * self.B,
            (1, 2): -5 * _sqrt3 * self.B,
            (1, 3): _sqrt2 * (self.B + self.C),
            (2, 3): -_sqrt6 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_3_1_states(self) -> Float64Array:
        """Calculate the E_3_1 states."""
        diag_elements = [
            +6 * self.Dq - 13 * self.B + 4 * self.C,
            +6 * self.Dq - 10 * self.B + 4 * self.C,
            -4 * self.Dq - 11 * self.B + 4 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -4 * self.B,
            (0, 2): 0.0,
            (1, 2): -_3sqrt2 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_3_2_states(self) -> Float64Array:
        """Calculate the A_3_2 states."""
        diag_elements = [
            +6 * self.Dq - 8 * self.B + 4 * self.C,
            -4 * self.Dq - 2 * self.B + 7 * self.C,
        ]
        off_diag_elements = {(0, 1): -12 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def A_1_2_states(self) -> Float64Array:
        """Calculate the A_1_2 states."""
        diag_elements = [
            +6 * self.Dq - 12 * self.B + 6 * self.C,
            -4 * self.Dq - 3 * self.B + 6 * self.C,
        ]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        GS = np.array([-4 * self.Dq - 21 * self.B])

        T_5_2 = np.array([0], dtype=np.float64)

        E_5_1 = np.array([6 * self.Dq - 21 * self.B]) - GS

        A_3_1 = np.array([6 * self.Dq - 12 * self.B + 4 * self.C]) - GS

        # Ligandfield dependent
        T_1_2 = -GS + self.T_1_2_states()
        T_3_1 = -GS + self.T_3_1_states()
        A_1_1 = -GS + self.A_1_1_states()
        E_1_1 = -GS + self.E_1_1_states()
        T_3_2 = -GS + self.T_3_2_states()
        T_1_1 = -GS + self.T_1_1_states()
        E_3_1 = -GS + self.E_3_1_states()
        A_3_2 = -GS + self.A_3_2_states()
        A_1_2 = -GS + self.A_1_2_states()

        if A_1_1[0] <= 1e-4:
            T_1_2 -= A_1_1[0]

            E_1_1 -= A_1_1[0]
            T_3_2 -= A_1_1[0]
            T_1_1 -= A_1_1[0]
            E_3_1 -= A_1_1[0]
            A_3_2 -= A_1_1[0]
            A_1_2 -= A_1_1[0]
            T_3_1 -= A_1_1[0]

            E_5_1 -= A_1_1[0]
            T_5_2 -= A_1_1[0]
            A_3_1 -= A_1_1[0]
            A_1_1 -= A_1_1[0]

        return {
            "3_T_1": T_3_1,
            "1_T_2": T_1_2,
            "1_A_1": A_1_1,
            "1_E_1": E_1_1,
            "3_T_2": T_3_2,
            "1_T_1": T_1_1,
            "3_E_1": E_3_1,
            "3_A_2": A_3_2,
            "1_A_2": A_1_2,
            "5_E_1": E_5_1,
            "5_T_2": T_5_2,
            "3_A_1": A_3_1,
        }


class d7(LigandFieldTheory):
    """Class for d7 configuration."""

    def __init__(self, Dq: float = 0.0, B: float = 971.0, C: float = 4499.0) -> None:
        """Initialize the d7 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def T_2_2_states(self) -> Float64Array:
        """Calculate the T_2_2 states."""
        diag_elements = [
            +12 * self.Dq + 5 * self.C,
            +2 * self.Dq - 6 * self.B + 3 * self.C,
            +2 * self.Dq + 4 * self.B + 3 * self.C,
            -8 * self.Dq + 6 * self.B + 5 * self.C,
            -8 * self.Dq - 2 * self.B + 3 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -_3sqrt3 * self.B,
            (0, 2): -5 * _sqrt3 * self.B,
            (0, 3): 4 * self.B + 2 * self.C,
            (0, 4): 2 * self.B,
            (1, 2): 3 * self.B,
            (1, 3): -_3sqrt3 * self.B,
            (1, 4): -_3sqrt3 * self.B,
            (2, 3): -_sqrt3 * self.B,
            (2, 4): +_sqrt3 * self.B,
            (3, 4): 10 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_2_1_states(self) -> Float64Array:
        """Calculate the T_2_1 states."""
        diag_elements = [
            +12 * self.Dq - 6 * self.B + 3 * self.C,
            +2 * self.Dq + 3 * self.C,
            +2 * self.Dq - 6 * self.B + 3 * self.C,
            -8 * self.Dq - 6 * self.B + 3 * self.C,
            -8 * self.Dq - 2 * self.B + 3 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -3 * self.B,
            (0, 2): +3 * self.B,
            (0, 3): 0.0,
            (0, 4): -_2sqrt3 * self.B,
            (1, 2): -3 * self.B,
            (1, 3): +3 * self.B,
            (1, 4): _3sqrt3 * self.B,
            (2, 3): -3 * self.B,
            (2, 4): -_sqrt3 * self.B,
            (3, 4): _2sqrt3 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_2_states(self) -> Float64Array:
        """Calculate the E_2 states."""
        diag_elements = [
            +12 * self.Dq - 6 * self.B + 3 * self.C,
            +2 * self.Dq + 8 * self.B + 6 * self.C,
            +2 * self.Dq - 1 * self.B + 3 * self.C,
            -18 * self.Dq - 8 * self.B + 4 * self.C,
        ]
        off_diag_elements = {
            (0, 1): -6 * _sqrt2 * self.B,
            (0, 2): -_3sqrt2 * self.B,
            (0, 3): 0.0,
            (1, 2): 10 * self.B,
            (1, 3): +_sqrt3 * (2 * self.B + self.C),
            (2, 3): _2sqrt3 * self.B,
        }
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_4_1_states(self) -> Float64Array:
        """Calculate the T_4_1 states."""
        diag_elements = [+2 * self.Dq - 3 * self.B, -8 * self.Dq - 12 * self.B]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, np.ndarray]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, np.ndarray]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states

        # Ligandfield multi dependent state become GS

        T_4_1 = self.T_4_1_states()

        # Ligendfield single dependent states

        GS = T_4_1[0]

        T_4_1[0] = np.array([0.0], dtype=float)
        A_4_2 = np.array([12 * self.Dq - 15 * self.B]) - GS
        T_4_2 = np.array([2 * self.Dq - 15 * self.B]) - GS

        A_2_1 = np.array([2 * self.Dq - 11 * self.B + 3 * self.C]) - GS
        A_2_2 = np.array([2 * self.Dq + 9 * self.B + 3 * self.C]) - GS

        # Ligandfield dependent
        T_2_2 = self.T_2_2_states() - GS
        T_2_1 = self.T_2_1_states() - GS
        E_2 = self.E_2_states() - GS
        T_4_1[1] -= GS

        if E_2[0] <= 0:
            A_4_2 -= E_2[0]
            T_4_2 -= E_2[0]
            A_2_1 -= E_2[0]
            A_2_2 -= E_2[0]
            T_2_2 -= E_2[0]
            T_2_1 -= E_2[0]
            T_4_1 -= E_2[0]
            E_2 -= E_2[0]

        return {
            "2_T_2": T_2_2,
            "2_T_1": T_2_1,
            "2_E": E_2,
            "4_T_1": T_4_1,
            "4_A_2": A_4_2,
            "4_T_2": T_4_2,
            "2_A_1": A_2_1,
            "2_A_2": A_2_2,
        }


class d8(LigandFieldTheory):
    """Class for d8 configuration."""
    def __init__(self, Dq: float = 0.0, B: float = 1030.0, C: float = 4850.0) -> None:
        """Initialize the d8 configuration with given parameters.

        Args:
            Dq (float): Crystal field splitting in wavenumbers (cm-1).
            B (float): Racah parameter B in wavenumbers (cm-1).
            C (float): Racah parameter C in wavenumbers (cm-1).

        """
        super().__init__(Dq, B, C)

    def A_1_1_states(self) -> Float64Array:
        """Calculate the A_1_1 states."""
        diag_elements = [
            +8 * self.Dq + 10 * self.B + 5 * self.C,
            -12 * self.Dq + 8 * self.B + 4 * self.C,
        ]
        off_diag_elements = {(0, 1): _sqrt6 * (2 * self.B + self.C)}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def E_1_states(self) -> Float64Array:
        """Calculate the E_1 states."""
        diag_elements = [+8 * self.Dq + self.B + 2 * self.C, -12 * self.Dq + 2 * self.C]
        off_diag_elements = {(0, 1): -_2sqrt3 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_1_2_states(self) -> Float64Array:
        """Calculate the T_1_2 states."""
        diag_elements = [+8 * self.Dq + self.B + 2 * self.C, -2 * self.Dq + 2 * self.C]
        off_diag_elements = {(0, 1): +_2sqrt3 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def T_3_1_states(self) -> Float64Array:
        """Calculate the T_3_1 states."""
        diag_elements = [+8 * self.Dq - 5 * self.B, -2 * self.Dq + 4 * self.B]
        off_diag_elements = {(0, 1): 6 * self.B}
        states = self.construct_matrix(diag_elements, off_diag_elements)
        return self.eigensolver(states)

    def solver(self) -> Dict[str, Float64Array]:
        """Solve for all states and return a dictionary of results.

        Returns:
            Dict[str, Float64Array]: Dictionary with atomic term symbols as keys and
                eigenvalues as values.

        """
        # Ligand field independent states

        # Ligendfield single depentent states

        GS = np.array([-12 * self.Dq - 8 * self.B])

        T_1_1 = np.array([-2 * self.Dq + 4 * self.B + 2 * self.C]) - GS
        T_3_2 = np.array([-2 * self.Dq - 8 * self.B]) - GS
        A_3_2 = np.array([0], dtype=float)
        # Ligandfield dependent
        A_1_1 = self.A_1_1_states() - GS
        E_1 = self.E_1_states() - GS
        T_1_2 = self.T_1_2_states() - GS
        T_3_1 = self.T_3_1_states() - GS

        return {
            "1_A_1": A_1_1,
            "1_E": E_1,
            "1_T_3": T_1_2,
            "3_T_1": T_3_1,
            "1_T_1": T_1_1,
            "3_T_2": T_3_2,
            "3_A_2": A_3_2,
        }
